/* $Id: dungeon.c,v 1.128 2010/03/12 12:54:18 oohara Exp $ */

#include <stdio.h>
/* realloc, abs, rand */
#include <stdlib.h>

#include "world.h"
#include "creature.h"
#include "list-int.h"
#include "util.h"
#include "creature-detail.h"
#include "grid-detail.h"
#include "scroll-detail.h"
#include "los.h"
#include "enchantment-detail.h"

#include "dungeon.h"

static int player_heal_by_kill(world *wp);
static int compare_inventory(const void *a, const void *b);
static int set_inventory_priority(world *wp, int how);

int
distance(int ax, int ay, int bx, int by)
{
  int dx;
  int dy;

  dx = bx - ax;
  dy = by - ay;

  if (abs(dx) > abs(dy))
    return abs(dx);
  
  return abs(dy);
}

int
compare_wait(list_int *p, list_int *q, void *data)
{
  creature *who_p;
  creature *who_q;
  world *wp = (world *) data;

  if (wp == NULL)
    return 0;

  who_p = wp->cr[p->n];
  who_q = wp->cr[q->n];

  if (who_p == NULL)
    return 1;
  if (who_q == NULL)
    return -1;

  if (who_p->wait < who_q->wait)
    return -1;
  if (who_p->wait > who_q->wait)
    return 1;

  return 0;
}

int
has_los_from_player(world *wp, int z, int x, int y)
{
  if (wp == NULL)
    return 0;

  if ((wp->player_id < 0) || (wp->player_id >= wp->cr_size))
    return 0;
  if (wp->cr[wp->player_id] == NULL)
    return 0;

  return has_los(wp, z, x, y,
                 wp->cr[wp->player_id]->z,
                 wp->cr[wp->player_id]->x,
                 wp->cr[wp->player_id]->y);
}

int
add_creature_here(world *wp, creature *who, int n)
{
  list_int *order = NULL;

  if (wp == NULL)
  {
    fprintf(stderr, "add_creature_here: wp is NULL\n");
    return -1;
  }
  if (who == NULL)
  {
    fprintf(stderr, "add_creature_here: who is NULL\n");
    return -1;
  }
  if (n < 0)
  {
    fprintf(stderr, "add_creature_here: strange n (%d)\n", n);
    return -1;
  }
  if (n == wp->player_id)
  {
    fprintf(stderr, "add_creature_here: n (%d) is for player\n", n);
    return -1;
  }

  while (n >= wp->cr_size)
  {
    if (world_expand_cr(wp) != 0)
    {  
      fprintf(stderr, "add_creature_here: world_expand_cr failed\n");
      return -1;
    }
  }

  if (wp->cr[n] != NULL)
  {
    fprintf(stderr, "add_creature_here: cr[%d] is already used\n", n);
    return -1;
  }

  order = list_int_new(n);
  if (order == NULL)
  {
    fprintf(stderr, "add_creature_here: list_int_new failed\n");
    return -1;
  }

  who->id = n;
  wp->cr[n] = who;
  wp->turn_order[who->z] = list_int_merge(wp->turn_order[who->z],
                                          order, (void *) wp,
                                          &compare_wait);
  return n;
}

int
add_creature(world *wp, creature *who)
{
  int i;

  if (wp == NULL)
  {
    fprintf(stderr, "add_creature: wp is NULL\n");
    return -1;
  }
  if (who == NULL)
  {
    fprintf(stderr, "add_creature: who is NULL\n");
    return -1;
  }

  for (i = 0; i < wp->cr_size; i++)
  {
    if ((i != wp->player_id) && (wp->cr[i] == NULL))
      break;
  }

  return add_creature_here(wp, who, i);
}

int
add_new_creature(world *wp, int type, int z, int x, int y, int wait)
{
  int i;
  int n;
  creature *who = NULL;

  if (wp == NULL)
  {
    fprintf(stderr, "add_new_creature: wp is NULL\n");
    return -1;
  }
  if (grid_is_illegal(wp, z, x, y))
  {
    fprintf(stderr, "add_new_creature: grid is illegal "
            "(%d, %d, %d)\n", z, x, y);
    return -1;
  }
  if (wait < 0)
  {
    fprintf(stderr, "add_new_creature: wait is negative (%d)\n", wait);
    return -1;
  }

  who = creature_new();
  if (who == NULL)
  {
    fprintf(stderr, "add_new_creature: creature_new failed\n");
    return -1;
  }

  who->id = -1;
  who->type = type;
  who->z = z;
  who->x = x;
  who->y = y;
  who->wait = wait;
  who->hp = 1;
  who->enchant_type = 0;
  who->enchant_duration = 0;
  for (i = 0; i < 4; i++)
    who->act[i] = 0;

  n = add_creature(wp, who);
  if (n < 0)
  {
    fprintf(stderr, "add_new_creature: add_creature failed\n");
    creature_delete(who);
    who = NULL;
    return -1;
  }

  creature_heal_hp(wp, n, -1, 0);

  return n;
}

int
remove_creature(world *wp, int creature_id)
{
  creature *who = NULL;
  list_int *order = NULL;
  list_int *order_temp = NULL;

  if (wp == NULL)
  {
    fprintf(stderr, "remove_creature: wp is NULL\n");
    return 1;
  }
  if ((creature_id < 0) || (creature_id >= wp->cr_size))
  {
    fprintf(stderr, "remove_creature: strange creature_id (%d)\n",
            creature_id);
    return 1;
  }
  if (wp->cr[creature_id] == NULL)
  {
    fprintf(stderr, "remove_creature: wp->cr[%d] is NULL\n", creature_id);
    return 1;
  }

  who = wp->cr[creature_id];

  if ((who->z < 0) || (who->z >= wp->size_z))
  {
    fprintf(stderr, "remove_creature: strange who->z (%d)\n", who->z);
    return 1;
  }

  order = wp->turn_order[who->z];
  while (order != NULL)
  {
    order_temp = order->next;
    if (order->n == who->id)
    {
      wp->turn_order[who->z] = list_int_unlink(order);
      list_int_delete(order);
      order = NULL;
    }
    order = order_temp;
  }
  wp->turn_order[who->z] = list_int_first(wp->turn_order[who->z]);

  wp->cr[creature_id] = NULL;
  who->id = -1;
  creature_delete(who);
  who = NULL;

  return 0;
}

int
creature_is_dead(world *wp, int creature_id)
{
  if (wp == NULL)
  {
    fprintf(stderr, "creature_is_dead: wp is NULL\n");
    return 1;
  }
  if ((creature_id < 0) || (creature_id >= wp->cr_size))
  {
    fprintf(stderr, "creature_is_dead: strange creature_id (%d)\n",
            creature_id);
    return 1;
  }

  if (wp->cr[creature_id] == NULL)
    return 1;
  if (wp->cr[creature_id]->hp <= 0)
    return 1;

  return 0;
}

int 
creature_is_friendly(world *wp, int creature_id)
{
  if (wp == NULL)
  {
    fprintf(stderr, "creature_is_friendly: wp is NULL\n");
    return 1;
  }
  if ((creature_id < 0) || (creature_id >= wp->cr_size))
  {
    fprintf(stderr, "creature_is_friendly: strange creature_id (%d)\n",
            creature_id);
    return 1;
  }

  if (creature_id == wp->player_id)
    return 1;
  if (creature_is_dead(wp, creature_id))
    return 0;
  switch (wp->cr[creature_id]->type)
  {
  case CR_PLAYER_SHADOW:
    return 1;
    break;
  default:
    break;
  }

  return 0;
}

int
creature_is_face_to_face(world *wp, int creature_id)
{
  creature *who = NULL;
  creature *player = NULL;

  if (wp == NULL)
  {
    fprintf(stderr, "creature_is_face_to_face: wp is NULL\n");
    return 0;
  }
  if ((creature_id < 0) || (creature_id >= wp->cr_size))
  {
    fprintf(stderr, "creature_is_face_to_face: strange creature_id (%d)\n",
            creature_id);
    return 0;
  }

  if (creature_id == wp->player_id)
    return 1;

  if (wp->candle <= 0)
    return 0;

  if (creature_is_dead(wp, creature_id))
    return 0;
  who = wp->cr[creature_id];
  if (wp->cr[wp->player_id] == NULL)
    return 0;
  player = wp->cr[wp->player_id];

  if (who->z != player->z)
    return 0;
  if (distance(who->x, who->y, player->x, player->y) > 6)
    return 0;
  if (!has_los(wp, who->z, who->x, who->y, player->z, player->x, player->y))
    return 0;

  return 1;
}

static int
player_heal_by_kill(world *wp)
{
  int i;

  if (wp == NULL)
    return 1;

  for (i = 0; i < wp->cr_size; i++)
  {
    if (creature_is_dead(wp, i))
      continue;
    if (!creature_is_friendly(wp, i))
      continue;

    creature_heal_hp(wp, i, -1, 0);
  }

  return 0;
}

int
creature_take_damage(world *wp, int target_id, int dam,
                     int attacker_id, int verbose)
{
  int target_is_friendly;
  int attacker_is_friendly;
  struct creature_detail crd_target;
  struct creature_detail crd_attacker;
  char buf[128];

  if (wp == NULL)
  {
    fprintf(stderr, "creature_take_damage: wp is NULL\n");
    return 1;
  }
  if ((target_id < 0) || (target_id >= wp->cr_size))
  {
    fprintf(stderr, "creature_take_damage: strange target_id (%d)\n",
            target_id);
    return 1;
  }
  if (dam <= 0)
  {
    fprintf(stderr, "creature_take_damage: dam is non-positive (%d)\n",
            dam);
    return 1;
  }

  if (creature_is_dead(wp, target_id))
    return 0;

  creature_detail_get(wp->cr[target_id]->type, &crd_target);
  /* creature_is_friendly() must be called before the damage is dealt
   * because a dead non-player creature can never be friendly
   */
  target_is_friendly = creature_is_friendly(wp, target_id);
  /* attacker_id can be negative */
  if ((attacker_id < 0) || (attacker_id >= wp->cr_size))
  {
    attacker_is_friendly = 0;
  }
  else
  {  
    attacker_is_friendly = creature_is_friendly(wp, attacker_id);
  }

  if (wp->cr[target_id]->enchant_duration > 0)
  {
    switch (wp->cr[target_id]->enchant_type)
    {
    case ENC_INVULNERABLE:
      return 0;
      break;
    default:
      break;
    }
  }

  wp->cr[target_id]->hp -= dam;

  if (verbose)
  {
    sprintf(buf, "%s is damaged", crd_target.name);
    add_log(wp, buf);
  }

  if (creature_is_dead(wp, target_id))
  {
    if ((attacker_id < 0)
        || (attacker_id >= wp->cr_size)
        || (wp->cr[attacker_id] == NULL))
    {
      sprintf(buf, "%s dies", crd_target.name);
      add_log(wp, buf);
    }
    else
    {
      creature_detail_get(wp->cr[attacker_id]->type, &crd_attacker);
      sprintf(buf, "%s kills %s",
              crd_attacker.name, crd_target.name);
      add_log(wp, buf);

      if ((attacker_is_friendly)
          && (!target_is_friendly))
      {
        player_heal_by_kill(wp);
      }
    }

    return 0;
  }

  switch (wp->cr[target_id]->type)
  {
  case CR_BERSERKER:
    creature_add_enchantment(wp, target_id,
                             ENC_FAST, 3);
    break;
  default:
    break;
  }

  return 0;
}

/* set amount (arg 3) to negative for full healing */
int
creature_heal_hp(world *wp, int target_id, int amount, int verbose)
{
  struct creature_detail crd;
  char buf[128];

  if (wp == NULL)
  {
    fprintf(stderr, "creature_heal_hp: wp is NULL\n");
    return 1;
  }
  if ((target_id < 0) || (target_id >= wp->cr_size))
  {
    fprintf(stderr, "creature_heal_hp: strange target_id (%d)\n",
            target_id);
    return 1;
  }

  if (creature_is_dead(wp, target_id))
    return 0;

  creature_detail_get(wp->cr[target_id]->type, &crd);

  if (amount < 0)
    wp->cr[target_id]->hp = crd.max_hp;
  else
    wp->cr[target_id]->hp += amount;

  if (wp->cr[target_id]->hp > crd.max_hp)
    wp->cr[target_id]->hp = crd.max_hp;

  if (verbose)
  {
    sprintf(buf, "%s feels better", crd.name);
    add_log(wp, buf);
  }

  return 0;
}

int
creature_count_this_type(world *wp, int type)
{
  int i;
  int n;

  if (wp == NULL)
  {
    fprintf(stderr, "creature_count_this_type: wp is NULL\n");
    return 0;
  }

  n = 0;
  for (i = 0; i < wp->cr_size; i++)
  {
    if (creature_is_dead(wp, i))
      continue;
    if (wp->cr[i]->type == type)
      n++;
  }

  return n;
}

int
creature_change_xy(world *wp, int creature_id, int x, int y)
{
  creature *who = NULL;
  if (wp == NULL)
  {
    fprintf(stderr, "creature_change_xy: wp is NULL\n");
    return 1;
  }
  if ((x < 0) || (x >= wp->size_x))
  {
    fprintf(stderr, "creature_change_xy: strange x (%d)\n", x);
    return 1;
  }
  if ((y < 0) || (y >= wp->size_y))
  {
    fprintf(stderr, "creature_change_xy: strange y (%d)\n", x);
    return 1;
  }

  if (wp->cr[creature_id] == NULL)
  {
    fprintf(stderr, "creature_change_xy: wp->cr[%d] is NULL\n",
            creature_id);
    return 1;
  }
  who = wp->cr[creature_id];

  if (who->id == wp->player_id)
  {
    wp->camera_x = x;
    wp->camera_y = y;
  }
  if ((who->z == wp->camera_z)
      && (who->x == wp->cursor_x)
      && (who->y == wp->cursor_y))
  {
    wp->cursor_x = x;
    wp->cursor_y = y;
  }
  who->x = x;
  who->y = y;

  return 0;
}

int
creature_change_z(world *wp, int creature_id, int z)
{
  creature *who = NULL;
  list_int *order = NULL;
  list_int *order_temp = NULL;

  if (wp == NULL)
  {
    fprintf(stderr, "creature_change_z: wp is NULL\n");
    return 1;
  }
  if ((creature_id < 0) || (creature_id >= wp->cr_size))
  {
    fprintf(stderr, "creature_change_z: strange creature_id (%d)\n",
            creature_id);
    return 1;
  }
  if ((z < 0) || (z >= wp->size_z))
  {
    fprintf(stderr, "creature_change_z: strange z (%d)\n", z);
    return 1;
  }

  if (wp->cr[creature_id] == NULL)
  {
    fprintf(stderr, "creature_change_z: wp->cr[%d] is NULL\n",
            creature_id);
    return 1;
  }
  who = wp->cr[creature_id];

  if (z == who->z)
    return 0;

  if (wp->turn_order[z] != NULL)
    who->wait += wp->cr[wp->turn_order[z]->n]->wait;

  order = wp->turn_order[who->z];
  while (order != NULL)
  {
    order_temp = order->next;
    if (order->n == who->id)
    {
      wp->turn_order[who->z] = list_int_unlink(order);
      wp->turn_order[z] = list_int_concat(wp->turn_order[z], order);
    }
    order = order_temp;
  }

  wp->turn_order[who->z] = list_int_sort(wp->turn_order[who->z],
                                         (void *) wp,
                                         &compare_wait);
  wp->turn_order[z] = list_int_sort(wp->turn_order[z],
                                    (void *) wp,
                                    &compare_wait);

  if (who->id == wp->player_id)
  {
    wp->turn_z = z;
    wp->camera_z = z;
    wp->camera_x = who->x;
    wp->camera_y = who->y;
    wp->cursor_x = who->x;
    wp->cursor_y = who->y;
  }
  who->z = z;

  return 0;
}

int
creature_teleport(world *wp, int creature_id)
{
  int i;
  int x;
  int y;
  int found;
  struct creature_detail crd;
  char buf[128];
  creature *who = NULL;

  if (wp == NULL)
  {
    fprintf(stderr, "creature_teleport: wp is NULL\n");
    return 1;
  }

  if (creature_is_dead(wp, creature_id))
    return 0;
  who = wp->cr[creature_id];

  found = 0;
  for (i = 0; i < 100; i++)
  {
    x = 1 + (rand() % (wp->size_x - 2));
    y = 1 + (rand() % (wp->size_y - 2));
    if ((i < 50)
        && (distance(x, y, who->x, who->y)))
      continue;
    if ((!grid_blocks_creature(wp, who->z, x, y))
        && (!grid_has_creature(wp, who->z, x, y)))
    {
      found = 1;
      break;
    }
  }

  if (!found)
  {
    add_log(wp, "teleport doesn't work in such a narrow space");
    return 0;
  }

  creature_detail_get(who->type, &crd);
  sprintf(buf, "%s teleports", crd.name);
  add_log(wp, buf);

  creature_change_xy(wp, who->id, x, y);

  return 0;
}

int
creature_add_enchantment(world *wp, int creature_id,
                         int type, int duration)
{
  struct creature_detail crd;
  char buf[128];
  creature *who = NULL;

  if (wp == NULL)
  {
    fprintf(stderr, "creature_add_enchantment: wp is NULL\n");
    return 1;
  }

  if (creature_is_dead(wp, creature_id))
    return 0;
  who = wp->cr[creature_id];
  creature_detail_get(who->type, &crd);

  if (who->enchant_duration > 0)
  {
    switch (who->enchant_type)
    {
    case ENC_ANTIMAGIC:
      if (type != ENC_ANTIMAGIC)
      { 
        sprintf(buf, "%s resists (%d) %s", crd.name, duration,
                enchantment_name(type));
        add_log(wp, buf);
        return 0;
      }
      break;
    default:
      break;
    }
  }

  if (who->enchant_type == type)
  {
    if (who->enchant_duration < duration)
      who->enchant_duration = duration;
    return 0;
  }

  sprintf(buf, "%s gets (%d) %s", crd.name, duration,
          enchantment_name(type));
  add_log(wp, buf);
  who->enchant_type = type;
  who->enchant_duration = duration;

  return 0;
}

int
add_object_here(world *wp, object *what, int n)
{
  if (wp == NULL)
  {
    fprintf(stderr, "add_object_here: wp is NULL\n");
    return -1;
  }
  if (what == NULL)
  {
    fprintf(stderr, "add_object_here: what is NULL\n");
    return -1;
  }
  if (n < 0)
  {
    fprintf(stderr, "add_object_here: strange n (%d)\n", n);
    return -1;
  }

  while (n >= wp->obj_size)
  {
    if (world_expand_obj(wp) != 0)
    {
      fprintf(stderr, "add_object_here: world_expand_obj failed\n");
      return -1;
    }
  }

  if (wp->obj[n] != NULL)
  {
    fprintf(stderr, "add_object_here: obj[%d] is already used\n", n);
    return -1;
  }

  what->id = n;
  wp->obj[n] = what;

  return n;
}

int
add_object(world *wp, object *what)
{
  int i;

  if (wp == NULL)
  {
    fprintf(stderr, "add_object: wp is NULL\n");
    return -1;
  }
  if (what == NULL)
  {
    fprintf(stderr, "add_object: what is NULL\n");
    return -1;
  }

  for (i = 0; i < wp->obj_size; i++)
  {
    if (wp->obj[i] == NULL)
      break;
  }

  return add_object_here(wp, what, i);
}


int
add_new_object(world *wp, int type, int z, int x, int y, int quantity)
{
  int n;
  object *what = NULL;

  if (wp == NULL)
  {
    fprintf(stderr, "add_new_object: wp is NULL\n");
    return -1;
  }
  if (grid_is_illegal(wp, z, x, y))
  {
    fprintf(stderr, "add_new_object: grid is illegal "
            "(%d, %d, %d)\n", z, x, y);
    return -1;
  }
  if (quantity < 0)
  {
    fprintf(stderr, "add_new_object: quantity is negative (%d)\n", quantity);
    return -1;
  }

  what = object_new();
  if (what == NULL)
  {
    fprintf(stderr, "add_new_object: object_new failed\n");
    return -1;
  }

  what->id = -1;
  what->type = type;
  what->z = z;
  what->x = x;
  what->y = y;
  what->quantity = quantity;

  n = add_object(wp, what);
  if (n < 0)
  {
    fprintf(stderr, "add_new_object: add_object failed\n");
    object_delete(what);
    what = NULL;
    return -1;
  }

  return n;
}

int
remove_object(world *wp, int object_id)
{
  object *what = NULL;

  if (wp == NULL)
  {
    fprintf(stderr, "remove_object: wp is NULL\n");
    return 1;
  }
  if ((object_id < 0) || (object_id >= wp->obj_size))
  {
    fprintf(stderr, "remove_object: strange object_id (%d)\n", object_id);
    return 1;
  }
  if (wp->obj[object_id] == NULL)
  {
    fprintf(stderr, "remove_object: wp->obj[%d] is NULL\n", object_id);
    return 1;
  }

  what = wp->obj[object_id];

  wp->obj[object_id] = NULL;
  what->id = -1;
  object_delete(what);
  what = NULL;

  return 0;
}

int
object_is_dead(world *wp, int object_id)
{
  if (wp == NULL)
  {
    fprintf(stderr, "object_is_dead: wp is NULL\n");
    return 1;
  }
  if ((object_id < 0) || (object_id >= wp->obj_size))
  {
    fprintf(stderr, "object_is_dead: strange object_id (%d)\n",
            object_id);
    return 1;
  }

  if (wp->obj[object_id] == NULL)
    return 1;
  if (wp->obj[object_id]->quantity <= 0)
    return 1;

  return 0;
}

int
scroll_is_unknown(world *wp, int type)
{
  if (wp == NULL)
  {
    fprintf(stderr, "scroll_is_unknown: wp is NULL\n");
    return 0;
  }

  if ((type >= 0) && (type < NUM_SCROLL) && (wp->label[type] >= 0))
    return 1;

  return 0;
}

char *
scroll_name(world *wp, int type)
{
  struct scroll_detail scd;
  char buf[128];

  if (wp == NULL)
  {
    fprintf(stderr, "scroll_name: wp is NULL\n");
    return NULL;
  }

  scroll_detail_get(type, &scd);

  if (scroll_is_unknown(wp, type))
  {
    sprintf(buf, "unknown %s #%d",
            (scd.healthy) ? "healthy" : "harmful",
            wp->label[type]);
    return concat_string(1, buf);
  }
  else
  {
    return concat_string(1, scd.real_name);
  }

  fprintf(stderr, "scroll_name: should not reach here\n");
  return NULL;
}

int
add_inventory_here(world *wp, int type, int quantity, int n)
{
  if (wp == NULL)
  {
    fprintf(stderr, "add_inventory_here: wp is NULL\n");
    return -1;
  }
  if (type < 0)
  {
    fprintf(stderr, "add_inventory_here: strange type (%d)\n", type);
    return -1;
  }
  if (quantity < 0)
  {
    fprintf(stderr, "add_inventory_here: quantity is negative (%d)\n",
            quantity);
    return -1;
  }

  while (n >= wp->inv_size)
  {
    if (world_expand_inv(wp) != 0)
    {
      fprintf(stderr, "add_inventory_here: world_expand_inv failed\n");
      return -1;
    }
  }

  if (wp->inv[n] == NULL)
  {
    fprintf(stderr, "add_inventory_here: wp->inv[%d] is NULL\n", n);
    return -1;
  }
  if ((wp->inv[n]->quantity > 0) && (wp->inv[n]->type != type))
  {
    fprintf(stderr, "add_inventory_here: inv[%d] is already used\n", n);
    return -1;
  }

  wp->inv[n]->type = type;
  wp->inv[n]->quantity += quantity;

  return n;
}

int
add_inventory(world *wp, int type, int quantity)
{
  int i;

  if (wp == NULL)
  {
    fprintf(stderr, "add_inventory: wp is NULL\n");
    return -1;
  }
  if (type < 0)
  {
    fprintf(stderr, "add_inventory: strange type (%d)\n", type);
    return -1;
  }
  if (quantity < 0)
  {
    fprintf(stderr, "add_inventory: quantity is negative (%d)\n",
            quantity);
    return -1;
  }

  sort_inventory(wp, 2);

  for (i = 0; i < wp->inv_size; i++)
  {
    if (wp->inv[i] == NULL)
    {
      fprintf(stderr, "add_inventory: wp->inv[i] is NULL\n");
      return -1;
    }
    if ((wp->inv[i]->quantity <= 0) || (wp->inv[i]->type == type))
      break;
  }

  return add_inventory_here(wp, type, quantity, i);
}

int
inventory_is_dead(world *wp, int inventory_id)
{
  if (wp == NULL)
  {
    fprintf(stderr, "inventory_is_dead: wp is NULL\n");
    return 1;
  }
  if ((inventory_id < 0) || (inventory_id >= wp->inv_size))
  {
    fprintf(stderr, "inventory_is_dead: strange inventory_id (%d)\n",
            inventory_id);
    return 1;
  }

  if (wp->inv[inventory_id]->type < 0)
    return 1;
  if (wp->inv[inventory_id]->quantity <= 0)
    return 1;

  return 0;
}

int
count_inventory(world *wp)
{
  int i;
  int length;

  if (wp == NULL)
  {
    fprintf(stderr, "count_inventory: wp is NULL\n");
    return 0;
  }

  length = 0;
  for (i = 0; i < wp->inv_size; i++)
  {
    if (!inventory_is_dead(wp, i))
      length++;
  }

  return length;
}

static int
compare_inventory(const void *a, const void *b)
{
  inventory * const *inv_a = (inventory * const *) a;
  inventory * const *inv_b = (inventory * const *) b;

  if ((*inv_b) == NULL)
    return -1;
  if ((*inv_a) == NULL)
    return 1;

  if (((*inv_b)->type < 0) || ((*inv_b)->quantity <= 0))
    return -1;
  if (((*inv_a)->type < 0) || ((*inv_a)->quantity <= 0))
    return 1;

  if ((*inv_a)->sort_priority > (*inv_b)->sort_priority)
    return -1;
  if ((*inv_a)->sort_priority < (*inv_b)->sort_priority)
    return 1;

  if ((*inv_a)->id < (*inv_b)->id)
    return -1;
  if ((*inv_a)->id > (*inv_b)->id)
    return 1;

  return 0;
}

/* how (arg 2):
 * 0: healthy first
 * 1: harmful first
 * 2: just check empty/non-empty and known/unknown
 */
static int
set_inventory_priority(world *wp, int how)
{
  int i;
  struct scroll_detail scd;

  if (wp == NULL)
    return 1;

  for (i = 0; i < wp->inv_size; i++)
  {
    if (wp->inv[i] == NULL)
      continue;

    wp->inv[i]->sort_priority = 0;

    if ((wp->inv[i]->type < 0) || (wp->inv[i]->quantity <= 0))
      continue;
    wp->inv[i]->sort_priority += 1;

    scroll_detail_get(wp->inv[i]->type, &scd);

    if (scroll_is_unknown(wp, wp->inv[i]->type))
      wp->inv[i]->sort_priority += 1;
    switch (how)
    {
    case 0:
      if (scd.healthy)
        wp->inv[i]->sort_priority += 2;
      break;
    case 1:
      if (!(scd.healthy))
        wp->inv[i]->sort_priority += 2;
      break;
    case 2:
      break;
    default:
      break;
    }
  }

  return 0;
}

int
sort_inventory(world *wp, int how)
{
  int i;

  if (wp == NULL)
  {
    fprintf(stderr, "sort_inventory: wp is NULL\n");
    return 1;
  }

  for (i = 0; i < wp->inv_size; i++)
  {
    if (wp->inv[i] == NULL)
      continue;
    wp->inv[i]->id = i;
  }
  set_inventory_priority(wp, how);
  qsort (wp->inv, wp->inv_size, sizeof(inventory *), &compare_inventory);
  for (i = 0; i < wp->inv_size; i++)
  {
    if (wp->inv[i] == NULL)
      continue;
    wp->inv[i]->id = i;
  }

  return 0;
}

/* return the slot number (which is non-negative) if found,
 * -1 on otherwise */
int
inventory_find_this_type(world *wp, int type)
{
  int i;

  if (wp == NULL)
  {
    fprintf(stderr, "inventory_find_this_type: wp is NULL\n");
    return -1;
  }

  for (i = 0; i < wp->inv_size; i++)
  {
    if (inventory_is_dead(wp, i))
      continue;
    if (wp->inv[i]->type == type)
      return i;
  }

  return -1;
}

int
grid_is_illegal(world *wp, int z, int x, int y)
{
  if (wp == NULL)
  {
    fprintf(stderr, "grid_is_illegal: wp is NULL\n");
    return 1;
  }

  if ((z < 0) || (z >= wp->size_z))
    return 1;
  if ((x < 0) || (x >= wp->size_x))
    return 1;
  if ((y < 0) || (y >= wp->size_y))
    return 1;

  return 0;
}

int
grid_blocks_creature(world *wp, int z, int x, int y)
{
  if (wp == NULL)
  {
    fprintf(stderr, "grid_block_creature: wp is NULL\n");
    return 1;
  }

  if (grid_is_illegal(wp, z, x, y))
    return 1;

  switch (wp->grid[z][x][y] & GRID_TERRAIN_MASK)
  {
  case GR_WALL:
  case GR_GLASS_WALL:
    return 1;
    break;
  default:
    break;
  }

  return 0;
}

int
grid_has_creature(world *wp, int z, int x, int y)
{
  int i;

  if (wp == NULL)
  {
    fprintf(stderr, "grid_has_creature: wp is NULL\n");
    return 1;
  }

  if (grid_is_illegal(wp, z, x, y))
    return 1;

  for (i = 0; i < wp->cr_size; i++)
  {
    if (wp->cr[i] == NULL)
      continue;
    if ((wp->cr[i]->z == z)
        && (wp->cr[i]->x == x)
        && (wp->cr[i]->y == y))
    {
      return 1;
    }
  }

  return 0;
}

int
grid_blocks_los(world *wp, int z, int x, int y)
{
  if (wp == NULL)
  {
    fprintf(stderr, "grid_blocks_los: wp is NULL\n");
    return 1;
  }

  if (grid_is_illegal(wp, z, x, y))
    return 1;

  switch (wp->grid[z][x][y] & GRID_TERRAIN_MASK)
  {
  case GR_WALL:
    return 1;
    break;
  default:
    break;
  }

  return 0;
}

int
grid_blocks_beam(world *wp, int z, int x, int y)
{
  if (wp == NULL)
  {
    fprintf(stderr, "grid_blocks_beam: wp is NULL\n");
    return 1;
  }

  if (grid_is_illegal(wp, z, x, y))
    return 1;

  switch (wp->grid[z][x][y] & GRID_TERRAIN_MASK)
  {
  case GR_WALL:
  case GR_GLASS_WALL:
    return 1;
    break;
  default:
    break;
  }

  return 0;
}

/* return 1 if the destination is found, 0 otherwise */
int
find_stair_destination(world *wp,
                       int from_z, int from_x, int from_y,
                       int *dest_z, int *dest_x, int *dest_y)
{
  int dz;
  int target;
  int z;
  int x;
  int y;

  if (wp == NULL)
  {
    fprintf(stderr, "find_stair_destination: wp is NULL\n");
    return 0;
  }
  if (grid_is_illegal(wp, from_z, from_x, from_y))
  {
    fprintf(stderr, "find_stair_destination: grid is illegal "
            "(%d, %d, %d)\n",
            from_z, from_x, from_y);
    return 0;
  }

  if (dest_z != NULL)
    *dest_z = from_z;
  if (dest_x != NULL)
    *dest_x = from_x;
  if (dest_y != NULL)
    *dest_y = from_y;

  dz = 0;
  target = -1;
  switch (wp->grid[from_z][from_x][from_y] & GRID_TERRAIN_MASK)
  {
  case GR_STAIR_DOWN_0:
    dz = 1;
    target = GR_STAIR_UP_0;
    break;
  case GR_STAIR_DOWN_1:
    dz = 1;
    target = GR_STAIR_UP_1;
    break;
  case GR_STAIR_UP_0:
    dz = -1;
    target = GR_STAIR_DOWN_0;
    break;
  case GR_STAIR_UP_1:
    dz = -1;
    target = GR_STAIR_DOWN_1;
    break;
  default:
    break;
  }

  if ((dz == 0) || (target < 0))
    return 0;

  z = from_z + dz;
  if ((z < 0) || (z >= wp->size_z))
  {
    fprintf(stderr, "find_stair_destination: stair points to outside\n");
    return 0;
  }
  for (x = 0; x < wp->size_x; x++)
  {
    for (y = 0; y < wp->size_y; y++)
    {
      if ((wp->grid[z][x][y] & GRID_TERRAIN_MASK) == target)
      {
        if (dest_z != NULL)
          *dest_z = z;
        if (dest_x != NULL)
          *dest_x = x;
        if (dest_y != NULL)
          *dest_y = y;
        return 1;
      }
    }
  }

  return 0;
}

int
candle_is_in_the_last(world *wp)
{
  if (wp == NULL)
  {
    fprintf(stderr, "candle_is_in_the_last: wp is NULL\n");
    return 0;
  }

  if ((wp->candle > 0) && (wp->candle <= 108))
    return 1;

  return 0;
}

int
add_log(world *wp, const char *str)
{
  char *s;

  if (wp == NULL)
  {
    fprintf(stderr, "add_log: wp is NULL\n");
    return 1;
  }

  if (str == NULL)
    return 0;

  s = concat_string(1, str);
  if (s == NULL)
  {
    fprintf(stderr, "add_log: concat_string failed\n");
    return 1;
  }

  if ((wp->log_head < 0) || (wp->log_head >= wp->log_size))
  {
    fprintf(stderr, "add_log: strange log_head (%d)\n", wp->log_head);
    free(s);
    s = NULL;
    return 1;
  }
  if (wp->log[wp->log_head] != NULL)
  {
    free(wp->log[wp->log_head]);
    wp->log[wp->log_head] = NULL;
  }
  wp->log[wp->log_head] = s;
  wp->log_head = (wp->log_head + 1) % wp->log_size;

  return 0;
}
